from __future__ import annotations
from dataclasses import dataclass
import re
from typing import Sequence

from frame_semantic_transformer.data.data_utils import standardize_punct

from .TaskSample import TaskSample
from .TriggerIdentificationTask import TriggerIdentificationTask


@dataclass
class TriggerIdentificationSample(TaskSample):
    task: TriggerIdentificationTask
    trigger_locs: list[int]

    # -- input / target for training --

    def get_target(self) -> str:
        output = ""
        prev_trigger_loc = 0
        for loc in sorted(self.trigger_locs):
            # TODO: handle these special chars better
            output += self.task.text[prev_trigger_loc:loc] + "*"
            prev_trigger_loc = loc
        output += self.task.text[prev_trigger_loc:]
        return standardize_punct(output)

    @staticmethod
    def evaluate_prediction(
        prediction_outputs: Sequence[str], target: str, _input: str
    ) -> tuple[int, int, int]:
        true_pos = 0
        false_pos = 0
        false_neg = 0

        prediction_parts = process_text_for_evaluation(prediction_outputs[0]).split()
        target_parts = process_text_for_evaluation(target).split()

        for i, target_part in enumerate(target_parts):
            pred_part = "" if i >= len(prediction_parts) else prediction_parts[i]
            is_target_trigger = target_part[0] == "*"
            is_pred_trigger = pred_part != "" and pred_part[0] == "*"
            target_content = target_part.replace("*", "")
            pred_content = pred_part.replace("*", "")
            # if the prediction is too short or the text doesn't match, the item is just wrong
            if target_content != pred_content:
                if is_target_trigger:
                    false_neg += 1
                else:
                    false_pos += 1
            elif is_target_trigger and is_pred_trigger:
                true_pos += 1
            elif is_target_trigger and not is_pred_trigger:
                false_neg += 1
            elif is_pred_trigger and not is_target_trigger:
                false_pos += 1

        # if the predictions are longer than the targets, then every extra item is just wrong
        if len(prediction_parts) > len(target_parts):
            false_pos += len(prediction_parts) - len(target_parts)

        return (true_pos, false_pos, false_neg)


def process_text_for_evaluation(sent: str) -> str:
    updated_sent = standardize_punct(sent)
    updated_sent = re.sub(r"\*\s+([a-zA-Z0-9])", r"*\1", updated_sent)
    updated_sent = re.sub(r"([a-zA-Z0-9])(\*?')", r"\1 \2", updated_sent)
    return updated_sent



__version__ = "0.3.2"

from .FrameSemanticTransformer import (
    FrameSemanticTransformer,
    DetectFramesResult,
    FrameElementResult,
    FrameResult,
)

__all__ = (
    "FrameSemanticTransformer",
    "DetectFramesResult",
    "FrameElementResult",
    "FrameResult",
)



from __future__ import annotations
from typing import Callable, Sequence

from .DataAugmentation import DataAugmentation


def chain_augmentations(
    augmentations: Sequence[DataAugmentation],
) -> Callable[[str, str], tuple[str, str]]:
    def chained_augmentation(input: str, output: str) -> tuple[str, str]:
        chained_input = input
        chained_output = output
        for augmentation in augmentations:
            chained_input, chained_output = augmentation(chained_input, chained_output)
        return chained_input, chained_output

    return chained_augmentation



from __future__ import annotations
from abc import ABC, abstractmethod
from random import uniform


class DataAugmentation(ABC):
    """
    Base class for data augmentations on training data
    """

    probability: float

    def __init__(self, probability: float):
        self.probability = probability

    def __call__(self, input: str, output: str) -> tuple[str, str]:
        """
        randomly apply this augmentation in proportion to self.probability
        """
        rand_val = uniform(0, 1.0)
        if rand_val > self.probability:
            return (input, output)
        return self.apply_augmentation(input, output)

    @abstractmethod
    def apply_augmentation(self, input: str, output: str) -> tuple[str, str]:
        """
        Main logic for subclasses to implement
        """
        pass



from __future__ import annotations
from .DataAugmentation import DataAugmentation
import re


def remove_contractions(text: str) -> str:
    new_text = text.replace("won't", "will not")
    new_text = new_text.replace("can't", "cannot")
    new_text = re.sub(r"n't(\b)", r" not\1", new_text)
    new_text = re.sub(r"'ll(\b)", r" will\1", new_text)
    new_text = re.sub(r"'m(\b)", r" am\1", new_text)
    new_text = re.sub(r"'re(\b)", r" are\1", new_text)
    new_text = re.sub(r"'ve(\b)", r" have\1", new_text)
    return new_text


class RemoveContractionsAugmentation(DataAugmentation):
    def apply_augmentation(self, input: str, output: str) -> tuple[str, str]:
        if "*'" in input or "*'" in output or "*n'" in input or "*n'" in output:
            return (input, output)

        return (
            remove_contractions(input),
            remove_contractions(output),
        )



from __future__ import annotations
from dataclasses import dataclass
from typing import Sequence
from frame_semantic_transformer.data.data_utils import standardize_punct
from frame_semantic_transformer.data.framenet import (
    get_core_frame_elements,
    get_non_core_frame_elements,
)

from .Task import Task


@dataclass
class ArgumentsExtractionTask(Task):
    text: str
    trigger_loc: int
    frame: str

    @staticmethod
    def get_task_name() -> str:
        return "args_extraction"

    def get_input(self) -> str:
        core_elements = get_core_frame_elements(self.frame)
        non_core_elements = get_non_core_frame_elements(self.frame)
        # put core elements in front
        elements = [*core_elements, *non_core_elements]
        return f"ARGS {self.frame} | {' '.join(elements)} : {self.trigger_labeled_text}"

    @staticmethod
    def parse_output(prediction_outputs: Sequence[str]) -> list[tuple[str, str]]:
        return split_output_fe_spans(prediction_outputs[0])

    @property
    def trigger_labeled_text(self) -> str:
        pre_span = self.text[0 : self.trigger_loc]
        post_span = self.text[self.trigger_loc :]
        # TODO: handle these special chars better
        return standardize_punct(f"{pre_span}*{post_span}")


def split_output_fe_spans(output: str) -> list[tuple[str, str]]:
    """
    Split an output like "Agent = He | Destination = to the store" into a list of elements and values, like:
    [("Agent", "He"), ("Destination", "to the store")]
    """
    outputs: list[tuple[str, str]] = []
    for span in output.split("|"):
        parts = span.strip().split("=")
        if len(parts) == 1:
            outputs.append((parts[0].strip(), "XXX"))
        else:
            outputs.append((parts[0].strip(), parts[1].strip()))
    return outputs



from __future__ import annotations
from collections import defaultdict
from typing import Any, Sequence, Type
from tqdm import tqdm
from transformers import T5ForConditionalGeneration, T5Tokenizer

from frame_semantic_transformer.data.data_utils import chunk_list
from frame_semantic_transformer.data.tasks.ArgumentsExtractionSample import (
    ArgumentsExtractionSample,
)
from frame_semantic_transformer.data.tasks.FrameClassificationSample import (
    FrameClassificationSample,
)
from frame_semantic_transformer.data.tasks.TaskSample import TaskSample
from frame_semantic_transformer.data.tasks.TriggerIdentificationSample import (
    TriggerIdentificationSample,
)
from frame_semantic_transformer.predict import batch_predict, predict_on_ids


def calc_eval_metrics(
    true_pos: float, false_pos: float, false_neg: float
) -> dict[str, float]:
    """
    Calculate precision, recall, and f score
    Based on https://github.com/swabhs/open-sesame/blob/master/sesame/evaluation.py
    """
    if true_pos == 0 and false_pos == 0:
        precision = 0.0
    else:
        precision = true_pos / (true_pos + false_pos)
    if true_pos == 0 and false_neg == 0:
        recall = 0.0
    else:
        recall = true_pos / (true_pos + false_neg)
    if precision == 0 and recall == 0:
        f_score = 0.0
    else:
        f_score = 2.0 * precision * recall / (precision + recall)
    return {"precision": precision, "recall": recall, "f_score": f_score}


def evaluate(
    model: T5ForConditionalGeneration,
    tokenizer: T5Tokenizer,
    samples: Sequence[TaskSample],
    batch_size: int = 10,
    print_failures: bool = False,
    predictions_per_sample: int = 5,
    top_k: int = 50,
    top_p: float = 0.95,
    repetition_penalty: float = 2.5,
    length_penalty: float = 1.0,
    early_stopping: bool = True,
    skip_special_tokens: bool = True,
    clean_up_tokenization_spaces: bool = True,
) -> dict[str, list[float]]:
    results: dict[str, list[float]] = defaultdict(lambda: [0, 0, 0])
    for samples_chunk in tqdm(
        chunk_list(samples, batch_size), total=len(samples) / batch_size
    ):
        inputs = [sample.get_input() for sample in samples_chunk]

        predictions = batch_predict(
            model,
            tokenizer,
            inputs,
            num_beams=predictions_per_sample,
            num_return_sequences=predictions_per_sample,
            top_k=top_k,
            top_p=top_p,
            repetition_penalty=repetition_penalty,
            length_penalty=length_penalty,
            early_stopping=early_stopping,
            skip_special_tokens=skip_special_tokens,
            clean_up_tokenization_spaces=clean_up_tokenization_spaces,
        )
        batched_predictions = chunk_list(predictions, predictions_per_sample)
        for sample, preds in zip(samples_chunk, batched_predictions):
            assert len(preds) == predictions_per_sample
            score = sample.evaluate_prediction(
                preds, sample.get_target(), sample.get_input()
            )
            true_pos, false_pos, false_neg = score
            results[sample.get_task_name()][0] += true_pos
            results[sample.get_task_name()][1] += false_pos
            results[sample.get_task_name()][2] += false_neg
            if print_failures and (false_neg > 0 or false_pos > 0):
                print(score)
                print(sample.get_target())
                print(preds)
                print("\n")

    return results


# TODO: figure out a better way to lookup the class from the string
TASK_SAMPLE_CLASS_MAP: dict[str, Type[TaskSample]] = {
    "args_extraction": ArgumentsExtractionSample,
    "frame_classification": FrameClassificationSample,
    "trigger_identification": TriggerIdentificationSample,
}


def evaluate_batch(
    model: T5ForConditionalGeneration,
    tokenizer: T5Tokenizer,
    batch: Any,
    predictions_per_sample: int = 5,
) -> dict[str, list[float]]:
    predictions = predict_on_ids(
        model,
        tokenizer,
        batch["input_ids"],
        batch["attention_mask"],
        skip_special_tokens=True,
        clean_up_tokenization_spaces=True,
        num_beams=predictions_per_sample,
        num_return_sequences=predictions_per_sample,
    )
    batched_predictions = chunk_list(predictions, predictions_per_sample)
    results: dict[str, list[float]] = defaultdict(lambda: [0, 0, 0])
    for preds, task, label, input_ids in zip(
        batched_predictions, batch["task"], batch["labels"], batch["input_ids"]
    ):
        assert len(preds) == predictions_per_sample
        target_tokens = [tok_id for tok_id in label.tolist() if tok_id != -100]
        input_tokens = [tok_id for tok_id in input_ids.tolist() if tok_id != -100]
        target = tokenizer.decode(
            target_tokens, skip_special_tokens=True, clean_up_tokenization_spaces=True
        )
        input = tokenizer.decode(
            input_tokens, skip_special_tokens=True, clean_up_tokenization_spaces=True
        )
        sample_class = TASK_SAMPLE_CLASS_MAP[task]
        true_pos, false_pos, false_neg = sample_class.evaluate_prediction(
            preds, target, input
        )
        results[task][0] += true_pos
        results[task][1] += false_pos
        results[task][2] += false_neg

    return results



from __future__ import annotations
from .DataAugmentation import DataAugmentation


class LowercaseAugmentation(DataAugmentation):
    def apply_augmentation(self, input: str, output: str) -> tuple[str, str]:
        task_def_index = input.find(":")
        task_def = input[:task_def_index]
        input_contents = input[task_def_index:]
        # only lowercase the content, not the task definition
        return (
            task_def + input_contents.lower(),
            output.lower(),
        )



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, Sequence


class Task(ABC):
    """
    Abstract interface for all Tasks
    """

    @staticmethod
    @abstractmethod
    def get_task_name() -> str:
        pass

    @staticmethod
    @abstractmethod
    def parse_output(prediction_outputs: Sequence[str]) -> Any:
        pass

    @abstractmethod
    def get_input(self) -> str:
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['frame_semantic_transformer',
 'frame_semantic_transformer.data',
 'frame_semantic_transformer.data.augmentations',
 'frame_semantic_transformer.data.tasks']

package_data = \
{'': ['*']}

install_requires = \
['nltk>=3.7,<4.0',
 'pytorch-lightning>=1.6.2,<2.0.0',
 'sentencepiece>=0.1.96,<0.2.0',
 'torch>=1.11.0,<2.0.0',
 'tqdm>=4.64.0,<5.0.0',
 'transformers>=4.18.0,<5.0.0']

setup_kwargs = {
    'name': 'frame-semantic-transformer',
    'version': '0.3.2',
    'description': 'Frame Semantic Parser based on T5 and FrameNet',
    'long_description': '# Frame Semantic Transformer\n[![ci](https://img.shields.io/github/workflow/status/chanind/frame-semantic-transformer/CI/main)](https://github.com/chanind/frame-semantic-transformer)\n[![PyPI](https://img.shields.io/pypi/v/frame-semantic-transformer?color=blue)](https://pypi.org/project/frame-semantic-transformer/)\n\n\nFrame-based semantic parsing library trained on [FrameNet](https://framenet2.icsi.berkeley.edu/) and built on HuggingFace\'s [T5 Transformer](https://huggingface.co/docs/transformers/model_doc/t5). This library is designed to be easy to use, yet powerful.\n\nLive Demo: [chanind.github.io/frame-semantic-transformer](https://chanind.github.io/frame-semantic-transformer)\n\nThis library draws heavily on [Open-Sesame](https://github.com/swabhs/open-sesame) ([paper](https://arxiv.org/abs/1706.09528)) for inspiration on training and evaluation on FrameNet, and uses ideas from the paper [Open-Domain Frame Semantic Parsing Using Transformers](https://arxiv.org/abs/2010.10998) for using T5 as a frame-semantic parser. [SimpleT5](https://github.com/Shivanandroy/simpleT5) was also used as a base for the initial training setup.\n\n## Installation\n\n```\npip install frame-semantic-transformer\n```\n\n## Usage\n\n### Inference\n\nThe main entry to interacting with the library is the `FrameSemanticTransformer` class, as shown below. For inference the `detect_frames()` method is likely all that is needed to perform frame parsing.\n\n```python\nfrom frame_semantic_transformer import FrameSemanticTransformer\n\nframe_transformer = FrameSemanticTransformer()\n\nresult = frame_transformer.detect_frames("The hallway smelt of boiled cabbage and old rag mats.")\n\nprint(f"Results found in: {result.sentence}")\nfor frames in result.frames:\n    print(f"FRAME: {frame.name}")\n    for element in frame.frame_elements:\n        print(f"{element.name}: {element.text}")\n```\n\nThe result returned from `detect_frames()` is an object containing `sentence`, a parsed version of the original sentence text, `trigger_locations`, the indices within the sentence where frame triggers were detected, and `frames`, a list of all detected frames in the sentence. Within `frames`, each object containes `name` which corresponds to the FrameNet name of the frame, `trigger_location` corresponding to which trigger in the text this frame this frame uses, and `frame_elements` containing a list of all relevant frame elements found in the text.\n\n### Loading Models\n\nThere are currently 2 available pre-trained models for inference, called `base` and `small`, fine-tuned from HuggingFace\'s [t5-base](https://huggingface.co/t5-base) and [t5-small](https://huggingface.co/t5-base) model respectively. If a local fine-tuned t5 model exists that can be loaded as well. If no model is specified, the `base` model will be used.\n\n```\nbase_transformer = FrameSemanticTransformer("base") # this is also the default\nsmall_transformer = FrameSemanticTransformer("small") # a smaller pretrained model which is faster to run\ncustom_transformer = FrameSemanticTransformer("/path/to/model") # load a custom t5 model\n```\n\nBy default, models are lazily loaded when `detect_frames()` is first called. If you want to load the model sooner, you can call `setup()` on a `FrameSemanticTransformer` instance to load models immediately.\n\n```\nframe_transformer = FrameSemanticTransformer()\nframe_transformer.setup() # load models immediately\n```\n\n### Training\n\nIf you want to train a new model yourself, you can run the training script like below:\n\n```\npython -m frame_semantic_transformer.train \\\n    --base-model t5-base \\\n    --use-gpu \\\n    --batch-size 8 \\\n    --epochs 10 \\\n    --learning-rate 1e-5 \\\n    --output-dir ./outputs\n```\n\nTraining uses [Pytorch Lightning](https://www.pytorchlightning.ai/) behind the scenes, and will place tensorboard logs into `./lightning_logs` as it trains.\n\nIf you need more control, you can also directly import the `train()` method from `frame_semantic_transformer.train` and run training directly in code.\n\n## Contributing\n\nAny contributions to improve this project are welcome! Please open an issue or pull request in this repo with any bugfixes / changes / improvements you have!\n\nThis project uses [Black](https://github.com/psf/black) for code formatting, [Flake8](https://flake8.pycqa.org/en/latest/) for linting, and [Pytest](https://docs.pytest.org/) for tests. Make sure any changes you submit pass these code checks in your PR. If you have trouble getting these to run feel free to open a pull-request regardless and we can discuss further in the PR.\n\n## License\n\nThe code contained in this repo is released under a MIT license, however the pretrained models are released under an Apache 2.0 license in accordance with FrameNet training data and HuggingFace\'s T5 base models.\n',
    'author': 'David Chanin',
    'author_email': 'chanindav@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/chanind/frame-semantic-transformer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations
from dataclasses import dataclass
from typing import Sequence
from frame_semantic_transformer.data.data_utils import standardize_punct
from frame_semantic_transformer.data.framenet import is_valid_frame

from .Task import Task


@dataclass
class FrameClassificationTask(Task):
    text: str
    trigger_loc: int

    # -- input / target for training --

    @staticmethod
    def get_task_name() -> str:
        return "frame_classification"

    def get_input(self) -> str:
        return f"FRAME: {self.trigger_labeled_text}"

    @staticmethod
    def parse_output(prediction_outputs: Sequence[str]) -> str | None:
        for pred in prediction_outputs:
            if is_valid_frame(pred):
                return pred
        return None

    # -- helper properties --

    @property
    def trigger_labeled_text(self) -> str:
        pre_span = self.text[0 : self.trigger_loc]
        post_span = self.text[self.trigger_loc :]
        # TODO: handle these special chars better
        return standardize_punct(f"{pre_span}*{post_span}")
